// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include "msg.h"
#include "furnace.h"
#include "vars.h"

Var heating("heating-mode");
Var cooling("cooling-mode");

static void
gotorc(int row, int column)
{
  printf("\033[%d;%dH", row, column);
}

static void
exit_proc(int)
{
  gotorc(23, 1);
  exit(0);
}

int twiddles[4], ftwiddle;
FurnaceData old_fd;

int
main(int argc, char **argv)
{
  Message *m;

  signal(SIGTERM, exit_proc);
  signal(SIGINT, exit_proc);

  if (argc > 1)
    Message::set_host(argv[1]);

  Message::set_my_name("console");
  Message::init();
  Var::wait_for_all_ids();

  setbuf(stdout, 0);
  printf("\033[1;1H\033[2J");

  while ((m = Message::next()) != NULL)
    {
      time_t now, flast;
      time(&now);
      struct tm *tm = localtime(&now);

      switch (m->type)
	{
#define MD(t,v) t v = *(t *)m->data
	case MT_climate:
	  {
	    static char *znames[] = {"MBR", "UPS", "DNS", "EXT"};
	    MD(ClimateData,cd);
	    gotorc(cd.zone+2, 2);
	    twiddles[cd.zone] = (twiddles[cd.zone] + 1) % 4;
	    printf("%c %s temp %6.2f  humid %5.1f ",
		   "|/-\\"[twiddles[cd.zone]],
		   znames[cd.zone], cd.temperature, cd.humidity);
	    gotorc(cd.zone+2, 40);
	    printf("%2d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
	  }
	  break;
	case MT_set_furnace:
	  {
	    int i;
	    MD(FurnaceData,fd);
	    gotorc(7, 2);
	    ftwiddle = (ftwiddle+1) % 4;
	    printf("%c %4s %4s %3s %5s ",
		   "|/-\\"[ftwiddle],
		   fd.heat ? "HEAT" : "----",
		   fd.cool ? "COOL" : "----",
		   fd.fan ? "FAN" : "---",
		   fd.humidify ? "HUMID" : "-----");
            if (cooling.value)
              printf("summer");
            else if (heating.value)
              printf("winter");
            else
              printf("-idle-");
	    if (memcmp(&old_fd, &fd, sizeof(fd)))
	      {
		old_fd = fd;
		flast = now;
		printf("  %2d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
	      }
	    for (i=0; i<4; i++)
	      {
		gotorc(i+2, 35);
		printf("%4s", fd.zone[i] ? "OPEN" : "    ");
	      }
	  }
	  break;
	}
      gotorc(9,11);
      printf("%2d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);
      gotorc(12,1);
    }
}
