// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "msg.h"
#include "furnace.h"
#include "logger.h"

#define EVENT_INTERVAL (5 * 60)

FurnaceData old_fd;
LogEvent le;
time_t last_event = 0;
time_t now, then;

int need = 0x3f;

void
check_dir(char *fname, int slashp)
{
  struct stat s;
  if (slashp)
    fname[slashp] = 0;
  if (stat(fname, &s))
    {
      if (mkdir(fname, 0777))
	{
	  fprintf(stderr, "error creating directory %s\n", fname);
	  perror("The error was");
	  exit(1);
	}
    }
  else
    {
      if (!S_ISDIR(s.st_mode))
	{
	  fprintf(stderr, "error, %s is not a directory\n", fname);
	  exit(1);
	}
    }
  if (slashp)
    fname[slashp] = '/';
}

void
write_event()
{
  if (need == 0x20)
    need = 0;
  if (need)
    return;

  le.time = now;

  struct tm *tm = localtime(&now);
  static char fname[100];
  sprintf(fname, "logs/%04d/%02d/%02d",
	  tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
  int fd = open(fname, O_WRONLY|O_APPEND|O_CREAT|O_SYNC, 0666);
  if (fd < 0)
    {
      check_dir(fname, 9);
      check_dir(fname, 12);
      fd = open(fname, O_WRONLY|O_APPEND|O_CREAT|O_SYNC, 0666);
    }
  if (fd < 0)
    {
      fprintf(stderr, "Error: cannot open log file %s for appending.\n");
      perror("The error was");
      return;
    }
  write(fd, &le, sizeof(le));
  close(fd);
}

int
main(int argc, char **argv)
{
  long timeout;
  Message *m;

  char *progdir = strdup(argv[0]);
  if (strrchr(progdir, '/'))
    {
      *strrchr(progdir, '/') = 0;
      chdir(progdir);
    }

  if (argc > 1)
    Message::set_host(argv[1]);

  check_dir("logs", 0);

  Message::set_my_name("logger");
  Message::init();

#define MD(t,v) t *v = (t *)m->data
	
  time(&now);

  while (1)
    {
      int z;

      then = last_event + EVENT_INTERVAL;
      timeout = then - now;
      if (timeout < 0)
	timeout = 0;

      m = Message::next(timeout * 1000);
	
      time(&now);

      if (!m)
	{
	  write_event();
	  last_event = now;
	}
      else
	switch (m->type)
	  {
	  case MT_climate:
	    {
	      MD(ClimateData,cd);
	      z = cd->zone;
	      le.temp[z] = (short int)(cd->temperature * 100);
	      le.humidity[z] = (unsigned char)(cd->humidity * 2);
	      need &= ~(1<<z);
	      if (need == 0x20)
		write_event();
	    }
	    break;
	  case MT_set_furnace:
	    {
	      MD(FurnaceData,fd);

	      if (memcmp(&old_fd, fd, sizeof(old_fd)) == 0)
		break;

	      need &= ~0x10;
	      old_fd = *fd;
	      le.furnace.heat = fd->heat;
	      le.furnace.cool = fd->cool;
	      le.furnace.fan = fd->fan;
	      le.furnace.humidify = fd->humidify;
	      le.furnace.zone0 = fd->zone[0];
	      le.furnace.zone1 = fd->zone[1];
	      le.furnace.zone2 = fd->zone[2];
	      le.furnace.zone3 = fd->zone[3];

	      write_event();
	    }
	    break;
	  }
    }
}
