// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>
#include <Xm/Scale.h>
#include <Xm/Frame.h>

#define MAXARGS 40
#define stdarg(p, v) XtSetArg(args[n], p, v), n++
int n;
Arg args[MAXARGS];

struct Process {
  Process *next;
  Process(char *cmd);
  const char *cmd;
  char **argv;
  int pid;
  int running;
  int keeprunning;
  Widget button;
  void setlabel(void);
  void start(void);
};

Process *process_root = 0;

Process::Process(char *Pcmd)
{
  keeprunning = 1;
  if (Pcmd[0] == '!')
    {
      keeprunning = 0;
      Pcmd++;
      while (Pcmd && isspace(*Pcmd))
	Pcmd++;
    }

  next = process_root;
  process_root = this;
  cmd = strdup(Pcmd);
  pid = -1;
  running = 0;
  button = 0;

  int argc = 2;
  char *c;
  for (c=Pcmd; *c; c++)
    if (isspace(c[0]) && isgraph(c[1]))
      argc++;

  Pcmd = strdup(Pcmd);
  argv = new char *[argc];
  argc = 0;
  argv[argc++] = Pcmd;
  for (c=Pcmd; *c; c++)
    {
      if (isspace(c[0]) && isgraph(c[1]))
	argv[argc++] = c+1;
      if (isspace(c[0]))
	c[0] = 0;
    }
  argv[argc] = 0;
}

void
Process::start(void)
{
  pid = fork();
  if (pid == 0)
    {
      execv (argv[0], argv);
      execvp (argv[0], argv);
      exit(0);
    }
  setlabel();
}

void
Process::setlabel(void)
{
  char label[110];
  switch (running * 2 + keeprunning)
    {
    case 0:
    case 1:
      sprintf(label, "S    -   %s ", cmd);
      break;
    case 2:
      sprintf(label, "X %6d %s ", pid, cmd);
      break;
    case 3:
      sprintf(label, "R %6d %s ", pid, cmd);
      break;
    }
  n = 0;
  XmString lstr = XmStringCreateLtoR(label, XmSTRING_DEFAULT_CHARSET);
  stdarg(XmNlabelString, lstr);
  XtSetValues(button, args, n);
  XFlush(XtDisplay(button));
}

void
sigchild_handler (int sig)
{
  int status;
  int pid = wait3(&status, WNOHANG, 0);
  Process *p;
  for (p=process_root; p; p=p->next)
    if (p->pid == pid)
      {
	if (p->running && p->keeprunning)
	  p->start();
	else
	  {
	    p->running = 0;
	    p->setlabel();
	  }
      }

  signal (SIGCHLD, sigchild_handler);
}

void
button_callback (Widget w, Process *p, XmPushButtonCallbackStruct *cb)
{
  if (p->running)
    {
      p->running = 0;
      if (p->pid > 0)
	kill (p->pid, SIGINT);
      p->pid = -1;
      p->setlabel();
    }
  else
    {
      p->running = 1;
      p->start();
    }
}

void
load_runlist (char *filename,
	      Widget form,
	      Widget above)
{
  FILE *in;
  char buf[100];
  char label[110];

  if (filename == 0)
    filename = "modrun.ini";
  in = fopen(filename, "r");
  if (!in)
    {
      perror(filename);
      exit(0);
    }

  while (fgets(buf, 100, in))
    {
      if (buf[0] == '#' || buf[0] <= ' ')
	continue;
      while (buf[strlen(buf)-1] <= ' ')
	buf[strlen(buf)-1] = 0;

      Process *p = new Process(buf);

      p->pid = -1;
      p->running = 0;

      n = 0;
      stdarg(XmNtopAttachment, XmATTACH_WIDGET);
      stdarg(XmNtopWidget, above);
      stdarg(XmNleftAttachment, XmATTACH_FORM);
      stdarg(XmNrightAttachment, XmATTACH_FORM);
      stdarg(XmNalignment, XmALIGNMENT_BEGINNING);
      p->button = XmCreatePushButton (form, p->argv[0], args, n);
      XtManageChild(p->button);
      XtAddCallback(p->button, XmNactivateCallback,
		    (XtCallbackProc)button_callback, p);

      p->setlabel();

      above = p->button;
    }
}

int
main(int argc, char **argv)
{

  Widget toplevel = XtInitialize(0, "XmModRun",
				 0, 0,
				 &argc, argv);


  n = 0;
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  stdarg(XmNshadowThickness, 0);
  Widget form_widget = XmCreateForm(toplevel, "form", args, n);
  XtManageChild(form_widget);

  n = 0;
  stdarg (XmNtopAttachment, XmATTACH_FORM);
  stdarg (XmNleftAttachment, XmATTACH_FORM);
  stdarg (XmNrightAttachment, XmATTACH_FORM);
  XmString lstr = XmStringCreateLtoR("Furnace Modules", XmSTRING_DEFAULT_CHARSET);
  stdarg(XmNlabelString, lstr);
  Widget title = XmCreateLabel(form_widget, "title", args, n);
  XtManageChild(title);

  load_runlist (argv[1], form_widget, title);

  signal (SIGCHLD, sigchild_handler);

  XtRealizeWidget(toplevel);
  XtMainLoop();
}
