// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <ctype.h>
#include <time.h>

#include "msg.h"
#include "furnace.h"

static void
setcolor (int nc)
{
  static int oldcolor = 0;
  if (nc != oldcolor)
    {
      oldcolor = nc;
      if (nc)
	printf("\033[%dm", 30+nc);
      else
	printf("\033[0m");
    }
}


static char *
ts(void)
{
  static char buf[40];
  time_t now;
  struct tm *t;
  time(&now);
  t = localtime(&now);
  sprintf(buf, "%02d:%02d:%02d", t->tm_hour, t->tm_min, t->tm_sec);
  return buf;
}

int
main(int argc, char **argv)
{
  Message *m;

  if (argc > 1)
    Message::set_host(argv[1]);

  Message::set_my_name("snoop");
  Message::set_snoop(1);
  Message::init();

  while ((m = Message::next()) != NULL)
    {
      int i;
      const char *tname;

      if (m->type < MT_MAX)
	tname = MsgTypeNames[m->type];
      else
	{
	  static char tbuf[10];
	  tname = tbuf;
	  sprintf(tbuf, "%d", m->type);
	}

      printf("%s %s.%d->%s.%d %s %d : ", ts(),
	     Message::get_name_for(m->source), m->source,
	     Message::get_name_for(m->dest), m->dest,
	     tname, m->length);

      for (i=0; i<m->length; i++)
	if (isgraph(m->data[i]))
	  {
	    setcolor(0);
	    putchar(m->data[i]);
	  }
	else
	  {
	    static int toggle = 0;
	    toggle = 1-toggle;
	    setcolor(toggle ? 3 : 6);
	    printf("%02x", m->data[i]);
	  }
      setcolor(0);
      printf("\n");

      switch (m->type)
	{
#define MD(t,v) t v = *(t *)m->data
	case MT_climate:
	  {
	    MD(ClimateData,cd);
	    printf("  zone %d temp %f humid %f\n", cd.zone, cd.temperature, cd.humidity);
	  }
	  break;
	}
    }
  return 0;
}
