// Copyright 2005 DJ Delorie
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <signal.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/time.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/RowColumn.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>
#include <Xm/Scale.h>
#include <Xm/Frame.h>

#include "msg.h"
#include "furnace.h"
#include "vars.h"

#include "eedigits.h"

#define MAXARGS 40
#define stdarg(p, v) XtSetArg(args[n], p, v), n++
int n;
Arg args[MAXARGS];

#define LCD_SCALE 2
#define LCD_MARGIN 2

Var last_ac_on("last-ac-on");

int monitor_mode = 0;

static unsigned char default_eeprom[17][480] = {
#include "eeprom.h"
};

struct LCD {
  Display *display;
  Window window;
  int screen;
  Colormap colormap;
  Widget widget, frame;
  GC gc0, gc1;
  int pix[4];
  unsigned char lcd[480];
  unsigned char eeprom[17][480];
  int backlight_on;
  int cursor_row, cursor_col;
  int font_page;
  int font_scale;
  int inverse_text;
  char buf[10];
  int bp, save_c;
  int numval, reading_number;
  int b_addr, x, y, x2, y2;
  int last_x, last_y;
  int prev_mask;

#define INK_BLACK 0x01
#define INK_XOR   0x02
  int ink_mode;
  int bottom_origin;

#define FWIDER  0x01
#define FTALLER 0x02
  int cw, ch, nc, nr;

  int io;

  void *state;
  int val;

  LCD(Widget w);
  void xinit(void);
  void reset(void);
  void redraw(int x, int y, int w, int h);
  static void lcd_callback(Widget, LCD *, XmDrawingAreaCallbackStruct *);

  void set_backlight(int b);
  void bs(void);
  void nl(void);
  void put_one(int c);
  void set_pixel(int x, int y);
  void draw_line(int x1, int y1, int x2, int y2);
  void reverse_lines(int new_mask);

  void write (int c);
  void write (char *c, int len=-1);
};

struct Button {
  int zone;
  int button;
  int pressed;
};

Button button[4][4];
double Temp[4];
double Humidity[4];
LCD *lcds[4];

Widget temp_scale[4];
Widget humidity_scale[4];

Pixel ion_fg, ion_bg, ioff_fg, ioff_bg;
Widget indicators[8];
FurnaceData saved_fd;

//-----------------------------------------------------------------------------

void
LCD::lcd_callback(Widget w, LCD *lcd, XmDrawingAreaCallbackStruct *cbs)
{
  int left, right, top, bottom;
  char c;
  KeySym keysym;

  switch (cbs->reason)
    {
    case XmCR_EXPOSE:
      if (!lcd->window)
	lcd->xinit();
      left = cbs->event->xexpose.x/LCD_SCALE;
      top = cbs->event->xexpose.y/LCD_SCALE;
      right = (cbs->event->xexpose.x + cbs->event->xexpose.width)/LCD_SCALE;
      bottom = (cbs->event->xexpose.y + cbs->event->xexpose.height)/LCD_SCALE;
      lcd->redraw(left, top, right-left+1, bottom-top+1);
      break;
    case XmCR_INPUT:
      switch (cbs->event->type)
	{
	case KeyPress:
	  if (XLookupString(&cbs->event->xkey, &c, 1, &keysym, 0) == 1)
	    lcd->write(c);
	  break;
	}
      break;
    case XmCR_RESIZE:
      break;
    }
}

LCD::LCD(Widget Pw)
{
  widget = Pw;
  display = XtDisplay(widget);
  window = XtWindow(widget);
  memcpy (eeprom, default_eeprom, sizeof(default_eeprom));

  reset();

  XtAddCallback(widget, XmNexposeCallback, (XtCallbackProc)LCD::lcd_callback, this);
  XtAddCallback(widget, XmNinputCallback, (XtCallbackProc)LCD::lcd_callback, this);
}

void
LCD::xinit(void)
{
#if 1
  static int colors[] = {70, 86, 30,
			 0, 0, 0,
			 168, 202, 24,
			 22, 48, 6};
#else
  static int colors[] = {168, 202, 24,
			 0, 0, 0,
			 255, 255, 255,
			 0, 0, 0 };
#endif
  window = XtWindow(widget);
  screen = XDefaultScreen(display);
  colormap = XDefaultColormap(display, screen);
  gc0 = XCreateGC(display, window, 0, 0);
  gc1 = XCreateGC(display, window, 0, 0);

  for (int i=0; i<4; i++)
    {
      XColor c;
      c.red   = colors[i*3] * 0x0101;
      c.green = colors[i*3+1] * 0x0101;
      c.blue  = colors[i*3+2] * 0x0101;
      c.flags = DoRed | DoGreen | DoBlue;
      if (XAllocColor (display, colormap, &c))
	pix[i] = c.pixel;
      else
	pix[i] = BlackPixel(display, screen);
    }

  set_backlight(0);
}

void
LCD::reset(void)
{
  memset(lcd, 0, sizeof(lcd));
  backlight_on = 0;
  cursor_row = cursor_col = 0;
  font_page = 0;
  font_scale = 0;
  inverse_text = 0;
  ink_mode = INK_BLACK;
  bottom_origin = 0;
  cw = 6;
  ch = 8;
  nc = 20;
  nr = 4;
  state = 0;
  reading_number = 0;
  b_addr = 0;
  prev_mask = 0;

  redraw(0, 0, 120, 32);
}

void
LCD::redraw(int x, int y, int w, int h)
{
  int xx, yy, p;
  if (!window)
    return;
  if (x+w > 120) w = 120-x;
  if (y+h > 32) h = 32-y;
  for (xx=x; xx<x+w; xx++)
    for (yy=y; yy<y+h; yy++)
      {
	p = lcd[xx+120*(yy/8)];
	p = (p >> (yy&7)) & 1;
	if (LCD_SCALE == 1)
	  XDrawPoint(display, window, p ? gc1 : gc0, xx, yy);
	else
	  XFillRectangle(display, window, p ? gc1 : gc0,
			 xx*LCD_SCALE, yy*LCD_SCALE,
			 LCD_SCALE, LCD_SCALE);
      }
}

void
LCD::set_backlight(int b)
{
  XSetForeground(display, gc0, pix[b*2]);
  XSetForeground(display, gc1, pix[b*2+1]);
  XSetWindowBorder(display, window, pix[b*2]);

  n = 0;
  stdarg(XmNbackground, pix[b*2]);
  XtSetValues(XtParent(widget), args, n);

  redraw(0, 0, 120, 32);
}

void
LCD::bs()
{
  int tc, tr;
  cursor_col --;
  if (font_scale & FWIDER && cursor_col & 1)
    cursor_col --;
  if (cursor_col < 0)
    {
      cursor_col = (font_scale & FWIDER ? 18 : 19);
      cursor_row += (font_scale & FTALLER ? 2 : 3);
      cursor_row %= 4;
    }
  tc = cursor_col;
  tr = cursor_row;
  put_one(' ');
  cursor_col = tc;
  cursor_row = tr;
}

void
LCD::nl(void)
{
  cursor_col = 0;
  cursor_row++;
  if (font_scale & FTALLER)
    if (cursor_row & 1)
      cursor_row++;
  cursor_row = cursor_row % 4;
}

void
LCD::put_one(int c)
{
  static unsigned char wider[] = {
    0x00, 0x03, 0x0c, 0x0f,
    0x30, 0x33, 0x3c, 0x3f,
    0xc0, 0xc3, 0xcc, 0xcf,
    0xf0, 0xf3, 0xfc, 0xff
  };
  int i, lp;
  unsigned char *font = eeprom[c < 112 ? font_page : font_page+1];
  if (c >= 112) c -= 112; else c -= 32;
  font += 6*c;
  if (cursor_col >= 20)
    nl();

  if (font_scale & FWIDER)
    cursor_col = (cursor_col+1) & ~1;
  if (font_scale & FTALLER)
    cursor_row = cursor_row & ~1;
  lp = cursor_row*120 + cursor_col * 6;

  for (i=0; i<6; i++)
    {
      if (font_scale & FTALLER)
	{
	  lcd[lp]     = wider[font[i]&0x0f] ^ inverse_text;
	  lcd[lp+120] = wider[font[i]/16] ^ inverse_text;
	  if (font_scale & FWIDER)
	    {
	      lp++;
	      lcd[lp]     = wider[font[i]&0x0f] ^ inverse_text;
	      lcd[lp+120] = wider[font[i]/16] ^ inverse_text;
	    }
	}
      else
	{
	  lcd[lp] = font[i] ^ inverse_text;
	  if (font_scale & FWIDER)
	    {
	      lcd[++lp] = font[i] ^ inverse_text;
	    }
	}
      lp++;
    }

  redraw(cursor_col*6, cursor_row*8, cw, ch);
  cursor_col ++;
  if (font_scale & FWIDER)
    cursor_col ++;
}

void
LCD::set_pixel(int x, int y)
{
  if (bottom_origin)
    y = 31-y;
  switch (ink_mode)
    {
    case 0:
    case INK_XOR:
      lcd[x+120*(y/8)] &= ~(1 << (y&7));
      break;
    case INK_BLACK:
      lcd[x+120*(y/8)] |= 1 << (y&7);
      break;
    case INK_BLACK | INK_XOR:
      lcd[x+120*(y/8)] ^= 1 << (y&7);
      break;
    }
  redraw(x, y, 1, 1);
}

void
LCD::draw_line(int x1, int y1, int x2, int y2)
{
  int lx, ly, dx, dy, r, i;
  x1 = x1%120;
  x2 = x2%120;
  y1 = y1%32;
  y2 = y2%32;

  last_x = x2;
  last_y = y2;

  lx = x2-x1;
  ly = y2-y1;
  dx = 1;
  dy = 1;

  if (lx < 0)
    {
      dx = -1;
      lx = -lx;
    }
  if (ly < 0)
    {
      dy = -1;
      ly = -ly;
    }
  if (lx > ly)
    {
      r = lx;
      lx *= 2;
      ly *= 2;
      for (i=0; i<lx; i+=2)
	{
	  set_pixel(x1, y1);
	  r += ly;
	  x1 += dx;
	  if (r > lx)
	    {
	      r -= lx;
	      y1 += dy;
	    }
	}
      set_pixel(x1, y1);
    }
  else
    {
      r = ly;
      lx *= 2;
      ly *= 2;
      for (i=0; i<ly; i+=2)
	{
	  set_pixel(x1, y1);
	  r += lx;
	  y1 += dy;
	  if (r > ly)
	    {
	      r -= ly;
	      x1 += dx;
	    }
	}
      set_pixel(x1, y1);
    }
}

void
LCD::reverse_lines(int new_mask)
{
  int i, c;
  int mask = new_mask ^ prev_mask;
  prev_mask = new_mask;

  for (i=0; i<4; i++)
    if (mask & (1<<i))
      {
	for (c=0; c<120; c++)
	  lcd[i*120+c] ^= 0xff;
	redraw(0, i*8, 120, 8);
      }
}

#define NEXTBYTE(lab) state = &&lab##_label; return; lab##_label:
#define DONEBYTE() state = 0; return
#define NUM(lab) reading_number = 1; numval = 0; NEXTBYTE(lab)

void
LCD::write(int c)
{
#if 0
  struct timeval tma, tmb;
  XFlush(display);
  gettimeofday(&tma, 0);
  while (1)
    {
      gettimeofday(&tmb, 0);
      if ((unsigned)(tmb.tv_usec - tma.tv_usec) >= 1000)
	break;
    }
#endif
  c &= 0xff;

  if (reading_number)
    {
      if (isdigit(c))
	{
	  reading_number = 2;
	  numval = numval * 10 + c - '0';
	  return;
	}
      else if (reading_number == 2)
	;
      else if (c >= 64 && c < 64*20*4)
	numval = c-64;
      reading_number = 0;
    }

  if (state)
    goto *state;

 unwrite:
  //  printf("top of write(), c = %c (%d)\n", isgraph(c) ? c : ' ', c);

  switch (c)
    {
    case 1:
      cursor_row = cursor_col = 0;
      break;

    case 2:
      inverse_text = ~0;
      break;

    case 3:
      inverse_text = 0;
      break;

    case 8:
      bs();
      break;

    case 9:
      cursor_col = (cursor_col + 4) & ~3;
      break;

    case 10:
      cursor_row++;
      if (font_scale & FTALLER)
	if (cursor_row && 1)
	  cursor_row++;
      cursor_row = cursor_row % 4;
      break;

    case 11:
      cursor_row--;
      if (font_scale & FTALLER)
	cursor_row--;
      cursor_row = (cursor_row + 4) % 4;
      break;

    case 12:
      cursor_row = cursor_col = 0;
      memset(lcd, 0, 480);
      redraw(0, 0, 120, 32);
      break;

    case 13:
      nl();
      break;

    case 14:
      set_backlight(1);
      break;

    case 15:
      set_backlight(0);
      break;

    case 16:
      NUM(L16);
      c = numval % (nc*nr);
      cursor_col = c % nc;
      cursor_row = c / nc;
      if (font_scale & FWIDER)
	cursor_col *= 2;
      if (font_scale & FTALLER)
	cursor_row *= 2;
      break;

    case 18:
      {
	int i;

	bp=0;
	NEXTBYTE(L18_1);
	if (c >= '2' && c <= '9')
	  {
	    c -= '0';
	    for (i=0; i<c; i++)
	      bs();
	    save_c = c;
	    while (1)
	      {
		NEXTBYTE(L18_2);
		if (c < 32 || c == '.' || bp == save_c)
		  goto unwrite;

		buf[bp++] = c;
	      }
	    for (i=0; i<(save_c-bp); i++)
	      put_one(' ');
	    for (i=0; i<bp; i++)
	      put_one(buf[i]);
	  }
      }
      break;

    case 27:
      {
	NEXTBYTE(L27);
	switch (c)
	  {
	  case 'A':
	    NUM(EAX);
	    x = numval;
	    NUM(EAY)
	    y = numval;
	    b_addr = x + y*120;
	    break;

	  case 'B':
	    b_addr %= 480;
	    NUM(EB);
	    lcd[b_addr] = numval;
	    redraw(b_addr%120, (b_addr/120)*8, 1, 8);
	    b_addr++;
	    break;

	  case 'D':
	    NEXTBYTE(ED);
	    if (c == 'G')
	      {
		for (x=0; x<480; x++)
		  {
		    NEXTBYTE(EDG);
		    lcd[x] = c;
		    redraw(x%120,(x/120)*8,1,8);
		  }
		//		redraw(0, 0, 120, 32);
	      }
	    break;

	  case 'E':
	    NUM(EE);
	    c = numval;
	    if (c >= 0 && c < 17)
	      {
		memcpy (lcd, eeprom[c], 480);
		redraw(0, 0, 120, 32);
	      }
	    break;

	  case 'F':
	    NUM(EF);
	    font_scale = numval & 3;
	    font_page = ((numval / 4)*2) % 16;
	    cw = (font_scale & FWIDER) ? 12 : 6;
	    ch = (font_scale & FTALLER) ? 16 : 8;
	    nr = 32/ch;
	    nc = 120/cw;
	    break;

	  case 'I':
	    NUM(EI);
	    if (numval & 1)
	      ink_mode |= INK_BLACK;
	    else
	      ink_mode &= ~INK_BLACK;
	    break;

	  case 'L':
	    NUM(EL1);
	    x = numval;
	    NUM(EL2);
	    y = numval;
	    NUM(EL3);
	    x2 = numval;
	    NUM(EL4);
	    y2 = numval;
	    draw_line(x, y, x2, y2);
	    break;

	  case 'M':
	    NUM(EM);
	    if (numval & 1)
	      ink_mode |= INK_XOR;
	    else
	      ink_mode &= ~INK_XOR;
	    break;

	  case 'P':
	    NUM(EPX)
	    x = numval % 120;
	    NUM(EPY);
	    y = numval % 32;
	    last_x = x;
	    last_y = y;
	    set_pixel(x, y);
	    break;

	  case 'R':
	    NUM(ER);
	    reverse_lines(numval);
	    break;

	  case 'T':
	    NUM(ET1);
	    x = numval;
	    NUM(ET2);
	    y = numval;
	    draw_line(last_x, last_y, x, y);
	    break;

	  case 'V':
	    NUM(EV);
	    bottom_origin = numval & 1;
	    break;

	  case 'W':
	    NUM(EW);
	    break;

	  case 'X':
	    NUM(EX);
	    if (numval >= 0 && numval < 17)
	      memmove (eeprom[numval], lcd, 480);
	    break;

	  }
      }

      break;

    default:
      if ((unsigned char)c >= 32 && (unsigned char)c <= 191)
	put_one(c);
      break;
    }

  DONEBYTE();
}

void LCD::write(char *c, int len)
{
  if (len == -1)
    len = strlen(c);
  while (len-- > 0)
    write(*c++);
}

//-----------------------------------------------------------------------------

static void
lprintf(int zone, char *fmt, ...)
{
  char buf[200];
  va_list a;

  if (zone < 0 || zone > 2)
    zone = 0;
  va_start (a, fmt);
  vsprintf(buf, fmt, a);
  lcds[zone]->write(buf);
  va_end(a);
}

static void
button_press (Widget w, Button *b, XmPushButtonCallbackStruct *pbcs)
{
  int newval = b->pressed;
  switch (pbcs->reason)
    {
    case XmCR_ACTIVATE:
    case XmCR_DISARM:
      newval = 0;
      break;
    case XmCR_ARM:
      newval = 1;
      break;
    }
  if (newval == b->pressed)
    return;
  b->pressed = newval;
  //  lprintf(b->zone, "\020@b z %d b %d pr %d ", b->zone, b->button, b->pressed);

  ButtonData bd;
  Message bdm (&bd, sizeof(bd), MT_button);
  bd.zone = b->zone;
  bd.button = b->button;
  bd.pressed = b->pressed;
  bdm.send();
}

static void
temp_cb (Widget w, double *v, XmScaleCallbackStruct *cbs)
{
  *v = (double)cbs->value / 10;
  //  lprintf(v-Temp, "\020TT[%d]: %5.1f F\n", v-Temp, *v);
}

static void
humidity_cb (Widget w, double *v, XmScaleCallbackStruct *cbs)
{
  *v = cbs->value;
  //  lprintf(v-Humidity, "\020hH[%d]: %3d   %%\n", v-Humidity, (int)*v);
}

static void
timer_cb ()
{
  static double oldtemp[4], oldhumidity[4];
  static int delay[4];
  int zone;
  for (zone=0; zone<4; zone++)
    {
      if (oldtemp[zone] != Temp[zone]
	  || oldhumidity[zone] != Humidity[zone]
	  || delay[zone]++ > 10)
	{

	  ClimateData cd;
	  cd.zone = zone;
	  cd.temperature = Temp[zone];
	  cd.humidity = Humidity[zone];
	  Message cdm (&cd, sizeof(cd), MT_climate);
	  cdm.send();

	  delay[zone] = 0;
	  oldtemp[zone] = Temp[zone];
	  oldhumidity[zone] = Humidity[zone];
	}

      if (zone < 3)
	{
	  int nzones = saved_fd.zone[0] + saved_fd.zone[1] + saved_fd.zone[2];
	  double t = Temp[zone];
	  if (saved_fd.heat && saved_fd.zone[zone])
	    t += (100-Temp[zone])/50 / nzones * (1+zone);
	  if (saved_fd.cool && saved_fd.zone[zone])
	    t += (50-Temp[zone])/30 / nzones * (1+zone);
	  t += (Temp[3] - Temp[zone])/100 * (5-zone)/5;
	  if (t != Temp[zone])
	    {
	      Temp[zone] = t;
	      XmScaleSetValue (temp_scale[zone], (int)(t * 10));
	    }

	  double h = Humidity[zone];
	  if ((saved_fd.humidify && saved_fd.zone[zone])
	      || h < Humidity[3])
	    h += 1.0;
	  else if (h > Humidity[3])
	    h -= 1.0;
	  if (h != Humidity[zone])
	    {
	      Humidity[zone] = h;
	      XmScaleSetValue (humidity_scale[zone], (int)h);
	    }
	}
    }

  XtAddTimeOut (300, (XtTimerCallbackProc)timer_cb, 0);
}

static Widget
create_zone (int zone, Widget form, Widget above)
{
  n = 0;
  if (above)
    {
      stdarg(XmNtopAttachment, XmATTACH_WIDGET);
      stdarg(XmNtopWidget, above);
    }
  else
    {
      stdarg(XmNtopAttachment, XmATTACH_FORM);
    }
  if (zone)
    stdarg(XmNtopOffset, 10);
  stdarg(XmNleftAttachment, XmATTACH_FORM);
  stdarg(XmNrightAttachment, XmATTACH_FORM);
  static char *znames[] = { "Master Bedroom", "Downstairs", "Upstairs", "Outside" };
  XmString lstr = XmStringCreateLtoR(znames[zone], XmSTRING_DEFAULT_CHARSET);
  stdarg(XmNlabelString, lstr);
  Widget header = XmCreateLabel(form, "header", args, n);
  XtManageChild(header);

  if (zone < 3)
    {
      n = 0;
      stdarg(XmNtopAttachment, XmATTACH_WIDGET);
      stdarg(XmNtopWidget, header);
      stdarg(XmNleftAttachment, XmATTACH_POSITION);
      stdarg(XmNleftPosition, 50);
      stdarg(XmNleftOffset, -(60*LCD_SCALE+LCD_MARGIN));
      stdarg(XmNwidth, 120*LCD_SCALE+2*LCD_MARGIN);
      stdarg(XmNheight, 32*LCD_SCALE+2*LCD_MARGIN);
      stdarg(XmNshadowThickness, 0);
      stdarg(XmNmarginHeight, LCD_MARGIN);
      stdarg(XmNmarginWidth, LCD_MARGIN);
      Widget lcdframe = XmCreateFrame(form, "lcdframe", args, n);
      XtManageChild(lcdframe);

      n = 0;
      Widget lcd = XmCreateDrawingArea(lcdframe, "lcd", args, n);
      XtManageChild(lcd);
      lcds[zone] = new LCD(lcd);

      Widget b[4];
      int i;
      for (i=0; i<4; i++)
	{
	  n = 0;
	  char *labels[4] = {"Up", "Down", "Left", "Right"};
	  XmString lstr = XmStringCreateLtoR(labels[i], XmSTRING_DEFAULT_CHARSET);
	  stdarg(XmNlabelString, lstr);
	  stdarg(XmNtopAttachment, XmATTACH_WIDGET);
	  stdarg(XmNtopWidget, lcdframe);
	  stdarg(XmNleftAttachment, XmATTACH_POSITION);
	  stdarg(XmNleftPosition, i*25);
	  stdarg(XmNrightAttachment, XmATTACH_POSITION);
	  stdarg(XmNrightPosition, i*25+25);
	  b[i] = XmCreatePushButton(form, "button", args, n);
	  XtManageChild(b[i]);

	  button[zone][i].zone = zone;
	  button[zone][i].button = i;
	  button[zone][i].pressed = 0;

	  XtAddCallback(b[i], XmNarmCallback,
			(XtCallbackProc)button_press, button[zone]+i);
	  XtAddCallback(b[i], XmNdisarmCallback,
			(XtCallbackProc)button_press, button[zone]+i);
	  XtAddCallback(b[i], XmNactivateCallback,
			(XtCallbackProc)button_press, button[zone]+i);
	}
      header = b[0];
    }

  n = 0;
  stdarg(XmNdecimalPoints, 1);
  stdarg(XmNminimum, 490);
  stdarg(XmNmaximum, 900);
  stdarg(XmNorientation, XmHORIZONTAL);
  stdarg(XmNshowValue, True);
  stdarg(XmNvalue, 700);
  Temp[zone] = 70.0;
  stdarg(XmNscaleMultiple, 10);
  stdarg(XmNtopAttachment, XmATTACH_WIDGET);
  stdarg(XmNtopWidget, header);
  stdarg(XmNleftAttachment, XmATTACH_FORM);
  stdarg(XmNrightAttachment, XmATTACH_POSITION);
  stdarg(XmNrightPosition, 60);
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  temp_scale[zone] = XmCreateScale(form, "temp", args, n);
  XtManageChild(temp_scale[zone]);
  XtAddCallback(temp_scale[zone], XmNdragCallback, (XtCallbackProc)temp_cb, Temp+zone);
  XtAddCallback(temp_scale[zone], XmNvalueChangedCallback, (XtCallbackProc)temp_cb, Temp+zone);

  n = 0;
  stdarg(XmNdecimalPoints, 0);
  stdarg(XmNminimum, 0);
  stdarg(XmNmaximum, 100);
  stdarg(XmNorientation, XmHORIZONTAL);
  stdarg(XmNshowValue, True);
  stdarg(XmNvalue, 35);
  Humidity[zone] = 35;
  stdarg(XmNtopAttachment, XmATTACH_WIDGET);
  stdarg(XmNtopWidget, header);
  stdarg(XmNleftAttachment, XmATTACH_POSITION);
  stdarg(XmNleftPosition, 60);
  stdarg(XmNrightAttachment, XmATTACH_FORM);
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  humidity_scale[zone] = XmCreateScale(form, "humidity", args, n);
  XtManageChild(humidity_scale[zone]);
  XtAddCallback(humidity_scale[zone], XmNdragCallback, (XtCallbackProc)humidity_cb, Humidity+zone);
  XtAddCallback(humidity_scale[zone], XmNvalueChangedCallback, (XtCallbackProc)humidity_cb, Humidity+zone);

  return humidity_scale[zone];
}

void
set_indicator(int which, int on)
{
  n = 0;
  stdarg(XmNforeground, on ? ion_fg : ioff_fg);
  stdarg(XmNbackground, on ? ion_bg : ioff_bg);
  XtSetValues(indicators[which], args, n);
}

Widget
create_furnace (Widget form, Widget above)
{
  n = 0;
  stdarg(XmNtopAttachment, XmATTACH_WIDGET);
  stdarg(XmNtopWidget, above);
  stdarg(XmNtopOffset, 10);
  stdarg(XmNleftAttachment, XmATTACH_FORM);
  stdarg(XmNrightAttachment, XmATTACH_FORM);
  char zname[10];
  sprintf(zname, "Furnace");
  XmString lstr = XmStringCreateLtoR(zname, XmSTRING_DEFAULT_CHARSET);
  stdarg(XmNlabelString, lstr);
  Widget header = XmCreateLabel(form, "header", args, n);
  XtManageChild(header);

  n = 0;
  stdarg(XmNtopAttachment, XmATTACH_WIDGET);
  stdarg(XmNtopWidget, header);
  stdarg(XmNleftAttachment, XmATTACH_FORM);
  stdarg(XmNrightAttachment, XmATTACH_FORM);
  stdarg(XmNshadowThickness, 0);
  Widget iform = XmCreateForm(form, "iform", args, n);
  XtManageChild(iform);

  int lefts[] = { 2, 26, 51, 76 };
  int rights[] = { 24, 49, 74, 98 };
  int tops[] = { 8, 53 };
  int bottoms[] = { 47, 92 };
  char *labels[8] = {"Heat", "Cool", "Fan", "Humid",
		     "Zone 1", "Zone 2", "Zone 3", "Zone 4"};
  int r, c;
  for (c=0; c<8; c++)
    {
      n = 0;
      stdarg(XmNshadowType, XmSHADOW_IN);
      stdarg(XmNshadowThickness, 2);
      stdarg(XmNtopAttachment, XmATTACH_POSITION);
      stdarg(XmNtopPosition, tops[c/4]);
      stdarg(XmNbottomAttachment, XmATTACH_POSITION);
      stdarg(XmNbottomPosition, bottoms[c/4]);
      stdarg(XmNleftAttachment, XmATTACH_POSITION);
      stdarg(XmNleftPosition, lefts[c%4]);
      stdarg(XmNrightAttachment, XmATTACH_POSITION);
      stdarg(XmNrightPosition, rights[c%4]);
      Widget iframe = XmCreateFrame(iform, "iframe", args, n);
      XtManageChild(iframe);

      n = 0;
      XmString lstr = XmStringCreateLtoR(labels[c], XmSTRING_DEFAULT_CHARSET);
      stdarg(XmNlabelString, lstr);
      indicators[c] = XmCreateLabel(iframe, "ind_off", args, n);
      XtManageChild(indicators[c]);
    }
  n = 0;
  Widget ion_w = XmCreateLabel(form, "ind_on", args, n);
  Widget ioff_w = XmCreateLabel(form, "ind_off", args, n);

  n = 0;
  stdarg(XmNforeground, &ion_fg);
  stdarg(XmNbackground, &ion_bg);
  XtGetValues(ion_w, args, n);
  n = 0;
  stdarg(XmNforeground, &ioff_fg);
  stdarg(XmNbackground, &ioff_bg);
  XtGetValues(ioff_w, args, n);

  set_indicator(1, 1);
  set_indicator(1, 0);
  set_indicator(2, 1);
  set_indicator(5, 1);
  return iform;
}

static void
xmp_read_message (void *ptr, int *source, XtInputId *id)
{
  Message *m;

  m = Message::next(0);
  if (!m)
    return;
  switch (m->type)
    {
    case MT_set_furnace:
      {
	FurnaceData *fd;
	fd = (FurnaceData *)m->data;
	set_indicator (0, fd->heat);
	set_indicator (1, fd->cool);
	set_indicator (2, fd->fan);
	set_indicator (3, fd->humidify);
	set_indicator (4, fd->zone[0]);
	set_indicator (5, fd->zone[1]);
	set_indicator (6, fd->zone[2]);
	set_indicator (7, fd->zone[3]);
	saved_fd = *fd;
      }
      break;

    case MT_ask_furnace:
      if (!monitor_mode)
	{
	  Message reply (&saved_fd, sizeof(saved_fd), MT_set_furnace);
	  reply.dest = m->source;
	  reply.send();
	}
      break;

    case MT_lcd:
      lcds[m->data[0]]->write((char *)m->data+1, m->length-1);
      break;

    case MT_climate:
      if (monitor_mode)
	{
	  ClimateData *cd = (ClimateData *)m->data;
	  XmScaleSetValue(temp_scale[cd->zone], (int)(cd->temperature * 10));
	  XmScaleSetValue(humidity_scale[cd->zone], (int)(cd->humidity));
	}
      break;
    }
}

static void
exit_proc(int)
{
  last_ac_on.set(0);
  exit(0);
}

int
main(int argc, char **argv)
{

  Widget toplevel = XtInitialize(0, "XmParallel",
				 0, 0,
				 &argc, argv);

  if (argc > 1)
    {
      Message::set_host(argv[1]);
      monitor_mode = 1;
      n = 0;
      stdarg(XmNtitle, argv[1]);
      XtSetValues(toplevel, args, n);
      argc--;
      argv++;
      memcpy (default_eeprom[0], eedigits[0], 480);
      memcpy (default_eeprom[1], eedigits[1], 480);
      memcpy (default_eeprom[2], eedigits[2], 480);
      memcpy (default_eeprom[3], eedigits[3], 480);
      memcpy (default_eeprom[4], eedigits[4], 480);
    }

  Message::set_my_name("parallel");
  Message::init();

  signal(SIGTERM, exit_proc);
  signal(SIGINT, exit_proc);


  n = 0;
  stdarg(XmNresizePolicy, XmRESIZE_ANY);
  stdarg(XmNshadowThickness, 0);
  Widget form_widget = XmCreateForm(toplevel, "form", args, n);
  XtManageChild(form_widget);

  Widget above = 0;
  for (int i=0; i<4; i++)
    above = create_zone(i, form_widget, above);

  create_furnace (form_widget, above);

  XtAddInput (Message::mux_socket(), (XtPointer)XtInputReadMask, (XtInputCallbackProc)xmp_read_message, 0);
  if (!monitor_mode)
    XtAddTimeOut (1000, (XtTimerCallbackProc)timer_cb, 0);

  XtRealizeWidget(toplevel);
  XtMainLoop();
}
